
const float PI = 3.14159265359;

layout(location = 0) out vec3 color;

in vec2 uv;

uniform sampler2D uWorldNormal;
uniform sampler2D uAlbedoRoughness;
uniform sampler2D uEmissive;
uniform sampler2D uDepth;

uniform sampler2D uShadowMap;
uniform samplerCube uRadiance;

uniform vec3 uLightColor;
uniform mat4 uLightMatrix;
uniform vec3 uCameraWorldPos;
uniform mat4 uProjMatrixInv;
uniform mat4 uViewMatrixInv;
uniform int uVolumetric;
uniform float uVolumetricSamples;

layout(std140, row_major) uniform Global
{
	float time;
} global;


float near_plane = 0.1;
float far_plane = 1000.0;

float LinearizeDepth(float depth)
{
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * near_plane * far_plane) / (far_plane + near_plane - z * (far_plane - near_plane));	
}

// clamps the dot product to 0.0
float sdot(vec3 v1, vec3 v2)
{
	return max(dot(v1, v2), 0.0);
}

// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

float ndfGGX(float LdotH, float roughness)
{
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;

	float denom = ( LdotH * LdotH ) * ( alpha2 - 1.0) + 1.0;
	return alpha2 / (PI * denom * denom);
}

// Single term for separable Schlick-GGX below.
float gaSchlickG1(float cosTheta, float k)
{
	return cosTheta / (cosTheta * (1.0 - k) + k);
}

// Schlick-GGX approximation of geometric attenuation function using Smith's method.
float gaSchlickGGX(float NdotL, float NdotV, float roughness)
{
	float r = roughness + 1.0;
	float k = (r * r) / 8.0;
	return gaSchlickG1(NdotL, k) * gaSchlickG1(NdotV, k);;
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;
	float NdotH = max(dot(N, H), 0.0);
	float NdotH2 = NdotH * NdotH;

	float num = alpha2;
	float denom = (NdotH2 * (alpha2 - 1.0) + 1.0);
	denom = PI * denom * denom;

	return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
	float r = (roughness + 1.0);
	float k = (r * r) / 8.0;

	float num = NdotV;
	float denom = NdotV * (1.0 - k) + k;

	return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
	float NdotV = max(dot(N, V), 0.0);
	float NdotL = max(dot(N, L), 0.0);
	float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}

float GetDirectionalLightOcclusion(vec3 location)
{
	// transform to normalized space
	vec4 lightPosition = uLightMatrix * vec4( location, 1.0);
	vec3 projLightCoords = lightPosition.xyz / lightPosition.w;
	projLightCoords = (projLightCoords * 0.5) + 0.5; // map from [-1, 1] to [0,1]

	if(projLightCoords.z > 1.0)
	{
		return  1.0f;// -0.0f;
	}

	float bias = 0.005;

	vec2 shadowMapTexelSize = 1.0 / textureSize(uShadowMap, 0);
	float current = projLightCoords.z;

	float shadowing = 0.0f;
	for(int x = -1; x <= 1; x++)
	{
		for(int y = -1; y <= 1; y++)
		{
			float depth = texture(uShadowMap, projLightCoords.xy + vec2( x, y) * shadowMapTexelSize ).r;
			shadowing += ((current - bias) > depth) ? 1.0 : 0.0;
		}
	}

	return 1.0 - (shadowing / 9.0);
}

// returns the world position form the depth buffer
// https://stackoverflow.com/questions/32227283/getting-world-position-from-depth-buffer-value
vec3 WorldPositionFromDepth(float depth, vec2 uv)
{
	float z = depth * 2.0 - 1.0; // form [0,1] to [-1, 1]
	
	vec4 clipSpace = vec4(uv * 2.0 - 1.0, z, 1.0); // clip space vertex
	vec4 viewSpace = uProjMatrixInv * clipSpace; // transform back to view space
	
	viewSpace /= viewSpace.w;
	
	return (uViewMatrixInv * viewSpace).xyz; // transform back to world position
}

// Henyey-Greenstein phase function
// https://bartwronski.files.wordpress.com/2014/08/bwronski_volumetric_fog_siggraph2014.pdf
// page 12
float ComputeScattering(float LdotV)
{
	float scattering = 0.7;
	float s2 = scattering * scattering;
	float result = 1.0 - s2;
	result /= (4.0 * PI * pow(1.0 + s2 - (2.0 * scattering) * LdotV, 1.5));
	return result;
}

// returns the uv coords in screen space [0,1]
vec2 getScreenSpaceUV()
{
	return gl_FragCoord.xy / textureSize( uDepth, 0);
}

void main()
{
	vec2 uv = getScreenSpaceUV();
	float depth = texture2D( uDepth, uv).r;

	if( depth == 1)
	{
		//color = vec3(1,0,0);
		//return;
		//discard;
	}

	vec3 worldPosition = WorldPositionFromDepth( depth, uv);

	vec3 LightVector =  vec3( 0.8, 1.0, 0.0);
	float distance = length(LightVector);

	vec3 normal = texture2D(uWorldNormal, uv).rgb;
	
	vec4 albedoRoughness = texture2D(uAlbedoRoughness, uv);
	float roughness = albedoRoughness.w;
	vec3 albedo = albedoRoughness.xyz;

	vec3 view = normalize(uCameraWorldPos - worldPosition);
	float NdotV = max(dot(normal, view), 0.0);
	vec3 reflection = 2.0 * NdotV * normal - view;

	vec3 lightDir = LightVector / distance;
	
	// half vector between view and light 
	vec3 half1 = normalize(view + lightDir);
	
	// 
	float NdotL = max(dot(normal, lightDir), 0.0);
	float NdotH = max(dot(normal, half1), 0.0);

	const float metallic = texture2D(uWorldNormal, uv).w;
	vec3 F0 = mix( vec3(0.04), albedo, metallic);

	vec3 F = fresnelSchlick( max( dot(half1, view), 0.0), F0);
	float D = ndfGGX(NdotH , roughness);
	float G = gaSchlickGGX(NdotL, NdotV, roughness);
	
	vec3 radiance = uLightColor * GetDirectionalLightOcclusion(worldPosition);  
	
	vec3 kd = mix(vec3(1) - F, vec3(0), metallic);
	
	vec3 diffuse = albedo * kd;
	vec3 specular = (F * D * G) / max(4.0 * NdotL * NdotV, 0.0001 );
	
	color = (diffuse + specular) * radiance * NdotL;
	
	// FOG
	if(uVolumetric == 0)
	{
		return;
	}

	vec3 viewRay = worldPosition - uCameraWorldPos;
	float rayLenght = length(viewRay);
	viewRay /= vec3(rayLenght);

	float stepLenght = rayLenght / 64.0;
	vec3 step = viewRay * stepLenght; 

	vec3 currentPosition = uCameraWorldPos;
	vec3 accumFog = vec3(0.0f);

	for(int i = 0; i < uVolumetricSamples; i++)
	{
		vec4 lightPosition = uLightMatrix * vec4( currentPosition, 1.0);
		vec3 projLightCoords = lightPosition.xyz / lightPosition.w;
		projLightCoords = (projLightCoords * 0.5) + 0.5;

		float shadow = texture(uShadowMap, projLightCoords.xy).r;
		if(shadow > projLightCoords.z )
		{
			accumFog += ComputeScattering(dot(viewRay, lightDir)) * uLightColor;
		}

		currentPosition += step;
	}



	accumFog /= uVolumetricSamples;
	color += accumFog;
}