
const float PI = 3.14159265359;

layout(location = 0) out vec3 color;

uniform sampler2D uWorldNormal;
uniform sampler2D uAlbedoRoughness;
uniform sampler2D uEmissive;
uniform sampler2D uDepth;

uniform vec3 uLightPos;
uniform float uLightRadius;
uniform vec3 uLightColor;
uniform vec3 uCameraWorldPos;
uniform mat4 uProjMatrixInv;
uniform mat4 uViewMatrixInv;

layout(std140, row_major) uniform Global
{
	float time;
} global;

float near_plane = 0.1;
float far_plane = 1000.0;

float LinearizeDepth(float depth)
{
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * near_plane * far_plane) / (far_plane + near_plane - z * (far_plane - near_plane));	
}

// clamps the dot product to 0.0
float sdot(vec3 v1, vec3 v2)
{
	return max(dot(v1, v2), 0.0);
}

// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

float ndfGGX(float LdotH, float roughness)
{
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;

	float denom = ( LdotH * LdotH ) * ( alpha2 - 1.0) + 1.0;
	return alpha2 / (PI * denom * denom);
}

// Single term for separable Schlick-GGX below.
float gaSchlickG1(float cosTheta, float k)
{
	return cosTheta / (cosTheta * (1.0 - k) + k);
}

// Schlick-GGX approximation of geometric attenuation function using Smith's method.
float gaSchlickGGX(float NdotL, float NdotV, float roughness)
{
	float r = roughness + 1.0;
	float k = (r * r) / 8.0;
	return gaSchlickG1(NdotL, k) * gaSchlickG1(NdotV, k);;
}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
	float alpha = roughness * roughness;
	float alpha2 = alpha * alpha;
	float NdotH = max(dot(N, H), 0.0);
	float NdotH2 = NdotH * NdotH;

	float num = alpha2;
	float denom = (NdotH2 * (alpha2 - 1.0) + 1.0);
	denom = PI * denom * denom;

	return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
	float r = (roughness + 1.0);
	float k = (r * r) / 8.0;

	float num = NdotV;
	float denom = NdotV * (1.0 - k) + k;

	return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
	float NdotV = max(dot(N, V), 0.0);
	float NdotL = max(dot(N, L), 0.0);
	float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}

// returns the world position form the depth buffer
// https://stackoverflow.com/questions/32227283/getting-world-position-from-depth-buffer-value
vec3 WorldPositionFromDepth(float depth, vec2 uv)
{
	float z = depth * 2.0 - 1.0; // form [0,1] to [-1, 1]
	
	vec4 clipSpace = vec4(uv * 2.0 - 1.0, z, 1.0); // clip space vertex
	vec4 viewSpace = uProjMatrixInv * clipSpace; // transform back to view space
	
	viewSpace /= viewSpace.w;
	
	return (uViewMatrixInv * viewSpace).xyz; // transform back to world position
}


// https://seblagarde.files.wordpress.com/2015/07/course_notes_moving_frostbite_to_pbr_v32.pdf
float smoothDistanceAtt(float squaredDistance, float invSqrAttRadius)
{
	float factor = squaredDistance * invSqrAttRadius;
	float smoothFactor = min(max(1.0 - factor * factor, 0.0), 1.0);
	return smoothFactor * smoothFactor;
}

// https://seblagarde.files.wordpress.com/2015/07/course_notes_moving_frostbite_to_pbr_v32.pdf
float getDistanceAtt(vec3 L, float invSqrAttRadius)
{
	float sqrtDist = dot(L, L);
	float attenuation = 1.0 / (max(sqrtDist, 0.01*0.01));
	attenuation *= smoothDistanceAtt(sqrtDist, invSqrAttRadius);
	return max(attenuation , 0.0);
}

// returns the uv coords in screen space [0,1]
vec2 getScreenSpaceUV()
{
	return gl_FragCoord.xy / textureSize( uDepth, 0);
}

void main()
{
	vec2 uv = getScreenSpaceUV();
	float depth = texture2D( uDepth, uv).r;

	if( depth == 1)
	{
		color = vec3(0.0);
		discard;
	}

	vec3 worldPosition = WorldPositionFromDepth( depth, uv);

	vec3 LightVector = uLightPos - worldPosition;
	float distance = length(LightVector);
	if(distance > uLightRadius)
	{
		color = vec3(0.0);
		discard;
	}

	vec3 normal = texture2D(uWorldNormal, uv).rgb;
	
	vec4 albedoRoughness = texture2D(uAlbedoRoughness, uv);
	float roughness = albedoRoughness.w;
	vec3 albedo = albedoRoughness.xyz;

	vec3 view = normalize(uCameraWorldPos - worldPosition);
	float NdotV = max(dot(normal, view), 0.0);
	vec3 reflection = 2.0 * NdotV * normal - view;

	vec3 lightDir = LightVector / distance;
	
	// half vector between view and light 
	vec3 half1 = normalize(view + lightDir);
	
	// 
	float NdotL = max(dot(normal, lightDir), 0.0);
	float NdotH = max(dot(normal, half1), 0.0);

	const float metallic = texture2D(uWorldNormal, uv).w;
	vec3 F0 = mix( vec3(0.04), albedo, metallic);

	vec3 F = fresnelSchlick( max( dot(half1, view), 0.0), F0);
	float D = ndfGGX(NdotH , roughness);
	float G = gaSchlickGGX(NdotL, NdotV, roughness);
	
	float invSqrAttRadius = 1.0 / (uLightRadius * uLightRadius);
	float attenuation = getDistanceAtt(LightVector, invSqrAttRadius);
	vec3 radiance = uLightColor * attenuation;  
	
	
	vec3 kd = mix(vec3(1) - F, vec3(0), metallic);
	
	vec3 diffuse = albedo * kd;
	vec3 specular = (F * D * G) / max(4.0 * NdotL * NdotV, 0.0001 );
	
	color = (diffuse + specular) * radiance * NdotL;
	
	// Translucency 
	if(metallic == 0.0)
	{
		float distortion = 0.1;
		float power = 4.0;
		float scale = 1.0;
		float thickness = 1.0;
		vec3 ambient = vec3(0);

		vec3 t =  lightDir + normal * distortion;

		//https://colinbarrebrisebois.com/2012/04/09/approximating-translucency-revisited-with-simplified-spherical-gaussian/
		float LdotT = exp2( clamp( dot(view, -t), 0.0, 1.0) * power - power) * scale;
		vec3 LT = attenuation * (LdotT + ambient) * thickness;
		color += albedo * uLightColor * max(LT, vec3(0.0));
	}
}